#ifndef __packet_h__
#define __packet_h__

#include <stdio.h>

#include "common.h"

struct IpHeader {
  uint8		ip_vhl;	/* version (4bits) & header length (4 bits) */
  uint8		ip_tos;	/* type of service */
  uint16	ip_len; /* length of IP datagram */
  uint16	ip_id;	/* identification (for frags) */ 
  uint16	ip_off; /* offset (within a fragment) and flags (3 bits) */
  uint8		ip_ttl; /* time to live */
  uint8		ip_proto;	/* protocol number */
  uint16	ip_xsum; /* checksum */
  uint32	ip_src; /* source address */
  uint32	ip_dst; /* destination address */
};

#define IP_DF 0x4000

/* Pseudo header for doing TCP checksum calculation */
struct PseudoIpHeader {
  uint32	filler[2];
  uint8		zero;
  uint8		ip_p;
  uint16	ip_len;
  uint32	ip_src;
  uint32	ip_dst;
};

struct tcpPseudoHdr {
  uint32 srcip;
  uint32 dstip;
  uint8 zero;
  uint8 ipproto;
  uint16 length;
} __attribute__((__packed__));


struct TcpHeader {
  uint16	tcp_sport;	/* source port */
  uint16	tcp_dport;	/* destination port */
  uint32	tcp_seq;	/* sequence number */
  uint32	tcp_ack;	/* acknoledgement number */
  uint8		tcp_hl;		/* header length (4 bits) */
  uint8		tcp_flags;	/* flags */
  uint16	tcp_win;	/* advertized window size */
  uint16	tcp_xsum;	/* checksum */
  uint16	tcp_urp;	/* urgent pointer */
};

struct UdpHeader {
  uint16 sourcePort;
  uint16 destPort;
  uint16 length;
  uint16 checksum;
};

class Packet {
protected:
  unsigned char payload[2000];
  int len;
  
  struct TcpHeader *findTcpHeader();
  struct UdpHeader *findUdpHeader();
  void recalculateChecksum();
  
public:
  const static int FLAG_FIN = 1<<0;
  const static int FLAG_SYN = 1<<1;
  const static int FLAG_RST = 1<<2;
  const static int FLAG_PSH = 1<<3;
  const static int FLAG_ACK = 1<<4;
  const static int FLAG_URG = 1<<5;
  const static int TCP_OPTION_MSS = 2;

  Packet();
  Packet(Packet *original);
  Packet(unsigned char *data, int len);
  unsigned char *getBuffer() { return payload; };
  int getBufferLen() { return sizeof(payload); };
  void setLength(int len) { this->len = len; };
  int getIpLength();
  bool isTcpPacket();
  bool isUdpPacket();
  Packet *clone();
  uint32 getSource() { return ntohl(((struct IpHeader*)&payload[0])->ip_src); };
  uint32 getDestination() { return ntohl(((struct IpHeader*)&payload[0])->ip_dst); };
  uint16 getSourcePort();
  uint16 getDestinationPort();
  bool isSource(struct sockaddr_in *addr);
  bool isDestination(struct sockaddr_in *addr);
  void copyFrom(unsigned char *data, int len);
  void dump(FILE *outfile);
  void setSourceAndDestination(struct sockaddr_in *source, struct sockaddr_in *destination);
  int getIpHeaderLength();
  int getTcpHeaderLength();
  int getControlBits() { return isTcpPacket() ? findTcpHeader()->tcp_flags : 0; };
  uint32 getTcpSequenceNo() { return isTcpPacket() ? ntohl(findTcpHeader()->tcp_seq) : 0; };
  uint32 getTcpAckNo() { return isTcpPacket() ? ntohl(findTcpHeader()->tcp_ack) : 0; };
  uint16 getIpId() { return ntohs(((struct IpHeader*)&payload[0])->ip_id); };
  uint8 getTTL() { return ((struct IpHeader*)&payload[0])->ip_ttl; };
  uint16 getTcpWindowSize() { return isTcpPacket() ? ntohs(findTcpHeader()->tcp_win) : 0; };
  uint16 getTcpUrgentPointer() { return isTcpPacket() ? ntohs(findTcpHeader()->tcp_urp) : 0; };
  uint8 *findTcpOption(int type);
  static uint16 checksum(uint8 *p1, int size1, uint8 *p2 = NULL, int size2 = 0);
  int getIpProtocolNumber() { return ((struct IpHeader*) &payload)->ip_proto; };
};
  
#endif /* defined(__packet_h__) */
