#include <string.h>
#include <math.h>

#include "monarch.h"
#include "probes.h"

//#define dprintf(a...) printf(a)
#define dprintf(a...) do { } while (0)

Diagnostics::Diagnostics(ProbeTrace *probeTrace)
{
  probeTrace->snapshot(&probesSent, &responsesReceived, &receiveSequence, &ipidSequence);
}

bool Diagnostics::ipidSanityCheck(char *msgBuf)
{
  /* Abort if we haven't received even a single response */

  if (responsesReceived < 1) {
    strcpy(msgBuf, "No responses received");
    return false;
  }

  /* Find out whether the IPID field is little endian or big endian. We add up the
     differences between consecutive IPIDs for each endianness, and then choose
     the the endianness with the lower total */

  bool ipidSeen[65536];
  for (int i=0; i<65536; i++)
    ipidSeen[i] = false;
  
  bool duplicateIpid = false;
  bool ipidAlwaysZero = true;
  int dist1=0, dist2=0;
  int lastId = -1;
  for (int i=0; i<responsesReceived; i++) {
    assert((0<=ipidSequence[i]) && (ipidSequence[i]<65536));
    if (ipidSequence[i] != 0)
      ipidAlwaysZero = false;
    if (ipidSeen[ipidSequence[i]])
      duplicateIpid = true;
    ipidSeen[ipidSequence[i]] = true;
    
    if (lastId < 0)
      lastId = ipidSequence[i];

    int d1 = ipidSequence[i] - lastId;
    if (d1 < -200) 
      d1 += 65536;
    if (d1 > 64000)
      d1 -= 65536;
    dist1 += d1;

    int d2 = ntohs(ipidSequence[i]) - ntohs(lastId);
    if (d2 < -200)
      d2 += 65536;
    if (d2 > 64000) 
      d2 -= 65536;
    dist2 += d2;
    
    lastId = ipidSequence[i];
  }
  
  /* Abort if the IPID field is always zero, or if we received any IPID twice */
  
  if (ipidAlwaysZero) {
    strcpy(msgBuf, "Remote host sets IPID field to zero");
    return false;
  }
  if (duplicateIpid) {
    strcpy(msgBuf, "Duplicate IPID detected");
    return false;
  }

  /* Some hosts randomize their IPIDs every once in a while. If this has happened, abort */

  double f1 = (1.0*dist1)/responsesReceived;
  double f2 = (1.0*dist2)/responsesReceived;
  double r = (fabs(f1)<fabs(f2)) ? f1 : f2;
  
  if ((r<0.5) || (r>4.0)) {
    sprintf(msgBuf, "Too many IPID gaps; r=%f", r);
    return false;
  }
  
  if (fabs(f2)<fabs(f1)) {
    for (int i=0; i<responsesReceived; i++)
      ipidSequence[i] = ntohs(ipidSequence[i]);
  }
  
  /* Some hosts don't use the LSB in the IPID field. If this applies here, 
     divide all IPIDs by two */
  
  bool usesLsb = false;
  for (int i=0; i<responsesReceived; i++) {
    if (ipidSequence[i] & 1)
      usesLsb = true;
  }

  if (!usesLsb) {
    for (int i=0; i<responsesReceived; i++)
      ipidSequence[i] >>= 1;
  }

  /* Determine the range of IPIDs we have seen, while paying attention
     to potential wrap-arounds */

  bool haveBelow10000 = false;
  int lowest = 65536, lowestAbove50000 = 65536;
  for (int i=0; i<responsesReceived; i++) {
    if (ipidSequence[i] < lowest)
      lowest = ipidSequence[i];
    if ((ipidSequence[i]>50000) && (ipidSequence[i] < lowestAbove50000))
      lowestAbove50000 = ipidSequence[i];
    if (ipidSequence[i] < 10000)
      haveBelow10000 = true;
  }
    
  if (haveBelow10000 && (lowestAbove50000 < 65536))
    lowest = lowestAbove50000;

  /* Normalize the IPIDs, so they start at zero */
    
  for (int i=0; i<responsesReceived; i++) {
    if (ipidSequence[i] < lowest)
      ipidSequence[i] += 65536;
    ipidSequence[i] -= lowest;
  }
   
  msgBuf[0] = 0; 
  return true;
}

#define MAX_SEQ_LEN 300

int m[MAX_SEQ_LEN+1][MAX_SEQ_LEN+1];

int Diagnostics::levenshteinDistance(int *a, int aLen, int *b, int bLen)
{
  if (!aLen)
    return bLen;
  if (!bLen)
    return aLen;

  if ((aLen > MAX_SEQ_LEN) || (bLen > MAX_SEQ_LEN))
    return -1;
    
  for (int ai=0; ai<=aLen; ai++)
    for (int bi=0; bi<=bLen; bi++)
      m[ai][bi] = (ai==0) ? bi : ((bi==0) ? ai : 0);
      
  bool madeProgress = true;
  while (madeProgress) {
    madeProgress = false;
    for (int ai=1; ai<=aLen; ai++) {
      for (int bi=1; bi<=bLen; bi++) {
        int min = 999;
        if ((ai>0) && ((m[ai-1][bi]+1)<min))
          min = m[ai-1][bi] + 1;
        if ((bi>0) && ((m[ai][bi-1]+1)<min))
          min = m[ai][bi-1] + 1; 
        if ((ai>0) && (bi>0)) {
          int cost = ((a[ai-1]==b[bi-1]) ? 0 : 1);
          if ((m[ai-1][bi-1]+cost)<min)
            min = m[ai-1][bi-1] + cost;
        }
        if (min != m[ai][bi]) {
          madeProgress = true;
          m[ai][bi] = min;
        }
      }
    }
  }
  
  return m[aLen][bLen];
}

void Diagnostics::condense(int *a, int aLen)
{
  for (int i=0; i<aLen; i++)
    assert(a[i]>=0);
  
  for (int i=0; i<aLen; i++) {
    int sti = 9999999;
    for (int j=0; j<aLen; j++)
      if ((a[j]>=i) && (a[j]<sti))
        sti = a[j];
    if (sti != i) {
      for (int j=0; j<aLen; j++)
        if (a[j]>=i)
          a[j] -= (sti-i);
    }
  }
}

void Diagnostics::findPermutation(int *seqA, int *seqB, int *perm, int len)
{
  int invPerm[len];
  for (int i=0; i<len; i++)
    invPerm[i] = -1;
    
  for (int i=0; i<len; i++) {
    assert((0<=seqA[i]) && (seqA[i]<len));
    invPerm[seqA[i]] = i;
  }
  
  for (int i=0; i<len; i++)
    assert(invPerm[i] >= 0);
    
  for (int i=0; i<len; i++) {
    assert((0<=seqB[i]) && (seqB[i]<len));
    perm[i] = invPerm[seqB[i]];
  }
}

void Diagnostics::write(FILE *stream)
{
  char buf1[256];
  
  if (!probesSent) {
    fprintf(stream, "No probe packets sent\n");
    return;
  }
  
  if (!responsesReceived) {
    fprintf(stream, "No responses received from remote host\n");
    return;
  }
  
  /* Abort if the IPID sequence is not sane */
  
  if (!ipidSanityCheck(buf1)) {
    fprintf(stream, "IPID sanity check fails (Reason: %s); cannot run self-diagnosis\n", buf1);
    return;
  }
  
  /* Append a 'guardian' packet at the end to make sure that the trace
     begins and ends with an in-sequence packet */
  
  int realIpid[responsesReceived+1], realReceiveSequence[responsesReceived+1];
  int highestIPID = 0, highestProbeSeq = 0;
  for (int i=0; i<responsesReceived; i++) {
    realReceiveSequence[i] = receiveSequence[i];
    if (receiveSequence[i] > highestProbeSeq)
      highestProbeSeq = receiveSequence[i];
    realIpid[i] = ipidSequence[i];
    if (ipidSequence[i] > highestIPID)
      highestIPID = ipidSequence[i];
  }
  
  assert(highestProbeSeq < probesSent);
  
  realReceiveSequence[responsesReceived] = probesSent;
  realIpid[responsesReceived] = highestIPID + 1;
  
  /* Find all in-sequence packets */
  
  int highestSeqBelow[responsesReceived+1], lowestSeqAbove[responsesReceived+1];
  int highestIpidBelow[responsesReceived+1], lowestIpidAbove[responsesReceived+1];

  highestSeqBelow[0] = -1;
  highestIpidBelow[0] = -1;
  for (int i=1; i<=responsesReceived; i++) {
    highestSeqBelow[i] = max(highestSeqBelow[i-1], realReceiveSequence[i-1]);
    highestIpidBelow[i] = max(highestIpidBelow[i-1], realIpid[i-1]);
  }

  assert(responsesReceived > 0);
  lowestSeqAbove[responsesReceived] = 9999999;
  lowestIpidAbove[responsesReceived] = 9999999;
  for (int i=responsesReceived-1; i>=0; i--) {
    lowestSeqAbove[i] = min(lowestSeqAbove[i+1], realReceiveSequence[i+1]);
    lowestIpidAbove[i] = min(lowestIpidAbove[i+1], realIpid[i+1]);
  }
    
  bool syncPoint[responsesReceived+1];
  for (int i=0; i<=responsesReceived; i++) {
    if ((highestSeqBelow[i] < realReceiveSequence[i]) && (highestIpidBelow[i] < realIpid[i]) &&
        (realReceiveSequence[i] < lowestSeqAbove[i]) && (realIpid[i] < lowestIpidAbove[i]))
      syncPoint[i] = true;
    else
      syncPoint[i] = false;
  }
    
  /* Some debug output */
    
  for (int i=0; i<=responsesReceived; i++) {
    dprintf("** R#%3d %s seq=%3d(%d) ipid=%d(%d) hSeqBelow=%d lSeqAbove=%d hIpidBelow=%d lIpidAbove=%d\n",
      i, syncPoint[i] ? "*" : " ", realReceiveSequence[i], (i>0) ? (realReceiveSequence[i]-realReceiveSequence[i-1]) : 0,
      realIpid[i], (i>0) ? (realIpid[i]-realIpid[i-1]) : 0,
      highestSeqBelow[i], lowestSeqAbove[i], highestIpidBelow[i], lowestIpidAbove[i]
    );
  }

  assert(syncPoint[0] && syncPoint[responsesReceived]);

  /* Look at each pair of in-sequence packets, and find out what happened (loss, reordering, cross traffic etc.) */

  int packetsLostDownstream = 0;
  int packetsLostUpstream = 0;
  int packetsCrosstraffic = 0;
  int reorderEpisodesDownstream = 0;
  int reorderEpisodesUpstreamMinor = 0;
  int reorderEpisodesUpstreamMajor = 0;
  int flowFeatures = 0;
  int p=0, lastSyncPoint = -1;
  while (p<=responsesReceived) {
    if (!syncPoint[p]) {
      p++;
      continue;
    }
        
    if (lastSyncPoint != -1) {
      dprintf("   E%3d-%3d: sent=%3d-%3d ipid=%3d-%3d ", lastSyncPoint, p, realReceiveSequence[lastSyncPoint], 
        realReceiveSequence[p], realIpid[lastSyncPoint], realIpid[p]
      );
        
      int packetsSent = realReceiveSequence[p] - realReceiveSequence[lastSyncPoint] - 1;
      int packetsReceived = p - lastSyncPoint - 1;
      int packetsLost = packetsSent - packetsReceived;
      int ipidsUsed = realIpid[p] - realIpid[lastSyncPoint] - 1;
      int ipidGaps = ipidsUsed - packetsReceived;
        
      dprintf("sent=%d rcvd=%d lost=%d ipid_gaps=%d", packetsSent, packetsReceived, packetsLost, ipidGaps);
        
      /* Detect cross traffic and losses */
        
      bool isUpstream = false, isDownstream = false;
      if (!packetsLost) {
        if (ipidGaps > 0) {
          flowFeatures |= FEATURE_CROSSTRAFFIC;
          packetsCrosstraffic += ipidGaps;
          dprintf(" -- %d packets crosstraffic", ipidGaps);
        }
        if (packetsSent > 0) {
          dprintf(" -- reordering");
        }
        dprintf("\n");
      } else if (ipidGaps == packetsLost) {
        dprintf(" -- %d upstream losses\n", packetsLost);
        packetsLostUpstream += packetsLost;
        flowFeatures |= FEATURE_LOSSES_UPSTREAM;
      } else if (ipidGaps == 0) {
        dprintf(" -- %d downstream losses\n", packetsLost);
        packetsLostDownstream += packetsLost;
        flowFeatures |= FEATURE_LOSSES_DOWNSTREAM;
      } else {
        dprintf(" -- %d unknown losses\n", packetsLost);
        flowFeatures |= FEATURE_UNKNOWN;
      }
        
      assert(realReceiveSequence[lastSyncPoint] < realReceiveSequence[p]);

      /* Detect reordering */

      if (packetsReceived > 0) {
        int aSeq[packetsReceived+2], bSeq[packetsReceived+2], cSeq[packetsReceived+2];
        for (int i=0; i<(packetsReceived+2); i++) {
          aSeq[i] = realReceiveSequence[lastSyncPoint+i] - realReceiveSequence[lastSyncPoint];
          bSeq[i] = realIpid[lastSyncPoint+i] - realIpid[lastSyncPoint];
          cSeq[i] = i;
        }
        
        /* Remove cross traffic and losses from the IPID sequence */
        
        condense(aSeq, packetsReceived+2);
        condense(bSeq, packetsReceived+2);

        /* There are three permutations we have to consider here: 
             1) downstream (local host to remote)
             2) upstream (remote back to local host)
             3) combined (what Monarch saw) */

        int dnPerm[packetsReceived+2], upPerm[packetsReceived+2], monarchPerm[packetsReceived+2];
        findPermutation(aSeq, bSeq, dnPerm, packetsReceived+2);
        findPermutation(bSeq, cSeq, upPerm, packetsReceived+2);
        findPermutation(aSeq, cSeq, monarchPerm, packetsReceived+2);
          
        dprintf("             reorder: snd[");
        for (int i=0; i<(packetsReceived+2); i++)
          dprintf("%s%d", (i==0) ? "" : " ", aSeq[i]);
        dprintf("] rmt[");
        for (int i=0; i<(packetsReceived+2); i++)
          dprintf("%s%d", (i==0) ? "" : " ", bSeq[i]);
        dprintf("] rcv[");
        for (int i=0; i<(packetsReceived+2); i++)
          dprintf("%s%d", (i==0) ? "" : " ", cSeq[i]);
        dprintf("]       dnPerm[");
        for (int i=0; i<(packetsReceived+2); i++)
          dprintf("%s%d", (i==0) ? "" : " ", dnPerm[i]);
        dprintf("] upPerm[");
        for (int i=0; i<(packetsReceived+2); i++)
          dprintf("%s%d", (i==0) ? "" : " ", upPerm[i]);
        dprintf("] monarchPerm[");
        for (int i=0; i<(packetsReceived+2); i++)
          dprintf("%s%d", (i==0) ? "" : " ", monarchPerm[i]);
        dprintf("]\n");

        /* How many dupACKs would Monarch have produced, vs TCP? */
          
        int atReceiverTCP[packetsReceived+2], atSenderTCP[packetsReceived+2];
        int atSenderMonarch[packetsReceived+2];
        int soFar = 0;
        for (int i=0; i<(packetsReceived+2); i++) {
          soFar = (dnPerm[i]>soFar) ? dnPerm[i] : soFar;
          atReceiverTCP[i] = soFar;
        }
        soFar = 0;
        for (int i=0; i<(packetsReceived+2); i++) {
          if (atReceiverTCP[upPerm[i]]>soFar)
            soFar = atReceiverTCP[upPerm[i]];
          atSenderTCP[i] = soFar;
        }
        soFar = 0;
        for (int i=0; i<(packetsReceived+2); i++) {
          soFar = (monarchPerm[i]>soFar) ? monarchPerm[i] : soFar;
          atSenderMonarch[i] = soFar;
        }
          
        dprintf("             ack_seq: rcv[");
        for (int i=0; i<(packetsReceived+2); i++)
          dprintf("%s%d", (i==0) ? "" : " ", atReceiverTCP[i]);
        dprintf("] snd[");
        for (int i=0; i<(packetsReceived+2); i++)
          dprintf("%s%d", (i==0) ? "" : " ", atSenderTCP[i]);
        dprintf("] monarch[");
        for (int i=0; i<(packetsReceived+2); i++)
          dprintf("%s%d", (i==0) ? "" : " ", atSenderMonarch[i]);
        dprintf("]\n");
         
        int dupackMaxTCP = 0, dupackCur;
        dupackCur = 0;
        for (int i=1; i<(packetsReceived+2); i++) {
          if (atSenderTCP[i] == atSenderTCP[i-1]) {
            dupackCur ++;
            if (dupackCur > dupackMaxTCP)
              dupackMaxTCP = dupackCur;
          } else {
            dupackCur = 0;
          }
        }
          
        int dupackMaxMonarch = 0;
        dupackCur = 0;
        for (int i=1; i<(packetsReceived+2); i++) {
          if (atSenderMonarch[i] == atSenderMonarch[i-1]) {
            dupackCur ++;
            if (dupackCur > dupackMaxMonarch)
              dupackMaxMonarch = dupackCur;
          } else {
            dupackCur = 0;
          }
        }
          
        dprintf("             dupack: TCP %d Monarch %d %s\n", dupackMaxTCP, dupackMaxMonarch, (dupackMaxTCP!=dupackMaxMonarch) ? "XXXXXX" : "");
        bool wasMajor = false;
        if (dupackMaxMonarch != dupackMaxTCP) {
          flowFeatures |= FEATURE_REORDERING_MESS;
          reorderEpisodesUpstreamMajor ++;
          wasMajor = true;
        }

        /* Find out where the reordering was, and set the appropriate flags */

        int l2 = levenshteinDistance(aSeq, packetsReceived+2, bSeq, packetsReceived+2);
        int l3 = levenshteinDistance(bSeq, packetsReceived+2, cSeq, packetsReceived+2);
        if (l2 > 0) {
          flowFeatures |= FEATURE_REORDERING_DOWNSTREAM;
          reorderEpisodesDownstream ++;
        }
        if (l3 > 0) {
          flowFeatures |= FEATURE_REORDERING_UPSTREAM;
          if (!wasMajor)
            reorderEpisodesUpstreamMinor ++;
        }
           
        dprintf("             here: dn=%d up=%d\n", l2, l3);
      }
    }

    lastSyncPoint = p++;
  }

  /* If we weren't able to classify some losses, we do not print any results, since they
     might be wrong (we could overestimate the downstream losses) */

  if (flowFeatures & FEATURE_UNKNOWN) {
    fprintf(stream, "Some losses cannot be classified as either downstream or upstream; cannot run self-diagnosis\n");
    return;
  }
  
  /* Print some statistics */
  
  fprintf(stream, "%3d probes sent\n", probesSent);
  fprintf(stream, "%3d responses received\n", responsesReceived);
  fprintf(stream, "%3d packets lost downstream\n", packetsLostDownstream);
  fprintf(stream, "%3d packets lost upstream\n", packetsLostUpstream);
  fprintf(stream, "%3d packets crosstraffic\n", packetsCrosstraffic);
  fprintf(stream, "%3d downstream reordering episodes\n", reorderEpisodesDownstream);
  fprintf(stream, "%3d minor upstream reordering episodes\n", reorderEpisodesUpstreamMinor);
  fprintf(stream, "%3d major upstream reordering episodes\n", reorderEpisodesUpstreamMajor);
  
  /* Print the overall verdict. Monarch is accurate if we have neither upstream losses
     nor upstream reordering. Certain types of upstream reordering do not make a 
     difference to TCP, so we report them separately. */
  
  fprintf(stream, "\nResult: ");
  if (!packetsLostUpstream && !reorderEpisodesUpstreamMajor) {
    if (!reorderEpisodesUpstreamMinor)
      fprintf(stream, "OK\n");
    else 
      fprintf(stream, "Minor upstream reordering\n");
  } else {
    if (packetsLostUpstream)
      fprintf(stream, "Upstream losses\n");
    else
      fprintf(stream, "Major upstream reordering\n");
  }
}
