#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <netdb.h>
#include <getopt.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>

#include "monarch.h"
#include "proxies.h"
#include "probes.h"
#include "traces.h"

volatile int senderPID = -1;
volatile int receiverPID = -1;
volatile bool shuttingDown = false;
volatile int exitCode = 99;
Proxy *proxy = NULL;
PacketTrace *packetTrace = NULL;
ProbeTrace *probeTrace = NULL;
KernelTrace *kernelTrace = NULL;
int transferSizeKB = -1;
int transferTimeSec = -1;
char packetTraceName[500];
char kernelTraceName[500];
char *messageFileName = NULL;
bool redirectInternalOutput = false;

static const char *procEntries[] = {
  "/proc/sys/kernel/ostype",
  "/proc/sys/kernel/osrelease",
  "/proc/sys/net/core/rmem_default",
  "/proc/sys/net/core/rmem_max",
  "/proc/sys/net/core/wmem_default",
  "/proc/sys/net/core/wmem_max",
  "/proc/sys/net/ipv4/tcp_adv_win_scale",
  "/proc/sys/net/ipv4/tcp_congestion_control",
  "/proc/sys/net/ipv4/tcp_bic",
  "/proc/sys/net/ipv4/tcp_westwood",
  "/proc/sys/net/ipv4/tcp_vegas_cong_avoid",
  "/proc/sys/net/ipv4/tcp_dsack",
  "/proc/sys/net/ipv4/tcp_ecn",
  "/proc/sys/net/ipv4/tcp_fack",
  "/proc/sys/net/ipv4/tcp_mem",
  "/proc/sys/net/ipv4/tcp_moderate_rcvbuf",
  "/proc/sys/net/ipv4/tcp_no_metrics_save",
  "/proc/sys/net/ipv4/tcp_rmem",
  "/proc/sys/net/ipv4/tcp_sack",
  "/proc/sys/net/ipv4/tcp_timestamps",
  "/proc/sys/net/ipv4/tcp_window_scaling",
  "/proc/sys/net/ipv4/tcp_wmem"
};

const char *columnName[] = {
  "state", "ca_state", "retransmits", "probes", "backoff", "options",
  "snd_wscale", "rcv_wscale", "rto", "ato", "snd_mss",
  "rcv_mss", "unacked", "sacked", "lost", "retrans", "fackets",
  "last_data_sent", "last_ack_sent", "last_dat_recv", "last_ack_recv", "pmtu", "rcv_ssthresh",
  "rtt", "rttvar", "snd_ssthresh", "snd_cwnd", "advmss", "reodering"
};

void usage(const char *programName)  
{
  printf("Usage: %s ", programName);
  printf("[-m mode] [-o packetTraceFile] [-k kernelTraceFile] [-M pathMTU]\n");
  printf("               [-z transferSizeKB] [-b bufferSizeKB] [-i interface]\n");
  printf("               [-t traceType] [-c captureMethod] [-p probeType] host:port\n");
  exit(ERR_PARAM);
}

void forkSenderAndReceiver(const char *progName, struct sockaddr_in *senderLocal, struct sockaddr_in *senderRemote, struct sockaddr_in *receiverLocal, struct sockaddr_in *receiverRemote, int socketBufferSizeKBytes)
{
  char buf1[256], buf2[256];

  /* Block the SIGUSR1 signal. The child process will send us this signal to indicate
     that it has bound to its port. Only then is it safe to start the sender. */

  sigset_t usr1;
  sigemptyset(&usr1);
  sigaddset(&usr1, SIGUSR1);
  sigprocmask(SIG_BLOCK, &usr1, 0);
  
  /* Create the receiver process */
  
  receiverPID = fork();
  if (receiverPID == 0) {
    if (redirectInternalOutput) {
      int out = open("receiver.stdout", O_CREAT | O_TRUNC | O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
      if (out < 0)
        panic(ERR_INTERNAL, "Cannot open receiver stdout redirect (%s)", strerror(errno));
      int err = open("receiver.stderr", O_CREAT | O_TRUNC | O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
      if (err < 0)
        panic(ERR_INTERNAL, "Cannot open receiver stderr redirect (%s)", strerror(errno));
      close(STDOUT_FILENO);
      dup2(out, STDOUT_FILENO);
      close(STDERR_FILENO);
      dup2(err, STDERR_FILENO); 
    }
    
    char cmdLine[1000];
    sprintf(cmdLine, "%s -m receiver --notify-parent -l %s", progName, renderAddress(receiverLocal, buf1));
    if (socketBufferSizeKBytes > 0)
      sprintf(&cmdLine[strlen(cmdLine)], " -b %d", socketBufferSizeKBytes);
    if (kernelTrace)
      sprintf(&cmdLine[strlen(cmdLine)], " -k %s", kernelTraceName);
    if (getVerbosity() == 1)
      sprintf(&cmdLine[strlen(cmdLine)], " -v");
    sprintf(&cmdLine[strlen(cmdLine)], " %s", renderAddress(receiverRemote, buf1));
      
    char *argv[50];
    argv[0] = strtok(cmdLine, " ");
    for (int i=1; i<50; i++)
      argv[i] = argv[i-1] ? strtok(NULL, " ") : NULL;
      
    execvp(progName, argv);
      
    panic(ERR_INTERNAL, "Cannot fork receiver process; exec failed (%s)", strerror(errno));
  }

  verbose(1, "Receiver process created (PID %d)", receiverPID);
  
  /* Wait for the SIGUSR1 signal */
  
  struct timespec timeout;
  timeout.tv_sec = 5;
  timeout.tv_nsec = 0;
  
  siginfo_t theSignal;
  int ret = sigtimedwait(&usr1, &theSignal, &timeout);
  if (ret != SIGUSR1)
    panic(ERR_INTERNAL, "Receiver has not sent SIGUSR1; maybe a problem with fork()? (ret=%d)", ret);
    
  /* Create the sender process */
    
  senderPID = fork();
  if (senderPID == 0) {
    if (redirectInternalOutput) {
      int out = open("sender.stdout", O_CREAT | O_TRUNC | O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
      if (out < 0)
        panic(ERR_INTERNAL, "Cannot open sender stdout redirect (%s)", strerror(errno));
      int err = open("sender.stderr", O_CREAT | O_TRUNC | O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
      if (err < 0)
        panic(ERR_INTERNAL, "Cannot open sender stderr redirect (%s)", strerror(errno));
      close(STDOUT_FILENO);
      dup2(out, STDOUT_FILENO);
      close(STDERR_FILENO);
      dup2(err, STDERR_FILENO); 
    }
    
    char cmdLine[1000];
    sprintf(cmdLine, "%s -m sender -l %s", progName, renderAddress(senderLocal, buf1));
    if(transferTimeSec >= 0)
      sprintf(&cmdLine[strlen(cmdLine)], " --transfer-time %d", transferTimeSec);
    else if (transferSizeKB >= 0)
      sprintf(&cmdLine[strlen(cmdLine)], " -z %d", transferSizeKB);
    if (socketBufferSizeKBytes > 0)
      sprintf(&cmdLine[strlen(cmdLine)], " -b %d", socketBufferSizeKBytes);
    if (messageFileName)
      sprintf(&cmdLine[strlen(cmdLine)], " --message-file %s", messageFileName);
    if (kernelTrace)
      sprintf(&cmdLine[strlen(cmdLine)], " -k %s", kernelTraceName);
    if (getVerbosity() == 1)
      sprintf(&cmdLine[strlen(cmdLine)], " -v");
    sprintf(&cmdLine[strlen(cmdLine)], " %s", renderAddress(senderRemote, buf1));
      
    char *argv[50];
    argv[0] = strtok(cmdLine, " ");
    for (int i=1; i<50; i++)
      argv[i] = argv[i-1] ? strtok(NULL, " ") : NULL;
      
    execvp(progName, argv);

    panic(ERR_INTERNAL, "Cannot fork receiver process; exec failed (%s)", strerror(errno));
  }

  verbose(1, "Sender process created (PID %d)", senderPID);
}

void handleChildExit(int code)
{
  /* If we started the internal sender and receiver processes, Monarch must terminate
     as soon as both processes have exited. Therefore, we listen for SIGCHLD signals
     and raise SIGTERM when we have seen both processes exit. */

  int status;
  pid_t child = wait(&status);
  if ((senderPID >= 0) || (receiverPID >= 0)) {
    if (child == senderPID) {
      verbose(1, "Sender process has terminated");
      senderPID = -1;
    }
    if (child == receiverPID) {
      verbose(1, "Receiver process has terminated");
      receiverPID = -1;
    }
    if ((senderPID<0) && (receiverPID<0) && !shuttingDown) {
      exitCode = 0;
      raise(SIGTERM);
    }
  }
  
  if ((status != 0) && !shuttingDown) {
    verbose(1, "Child process %d exited with code %d, terminating Monarch", child, status);
    raise(SIGTERM);
  }
}

void handleTermination(int code)
{
  if (shuttingDown)
    exit(exitCode);
    
  shuttingDown = true;

  /* Release any allocated resources, and write all traces. This may be called
     at any time (e.g. if some class panics during initialization), so not
     all resources may be allocated yet */

  if (senderPID >= 0) {
    verbose(1, "Terminating sender process (PID %d)", senderPID);
    int pid = senderPID;
    senderPID = -1;
    kill(pid, SIGTERM);
  }
  if (receiverPID >= 0) {
    verbose(1, "Terminating receiver process (PID %d)", receiverPID);
    int pid = receiverPID;
    receiverPID = -1;
    kill(pid, SIGTERM);
  }
  if (proxy) {
    verbose(1, "Shutting down %s proxy", proxy->getName());
    delete proxy;
    proxy = NULL;
  }
  if (packetTrace) {
    verbose(1, "Finalizing %s packet trace", packetTrace->getName());
    packetTrace->finalize();
    delete packetTrace;
    packetTrace = NULL;
  }
  if (probeTrace) {
    verbose(1, "Running internal diagnostics");
    Diagnostics *diagnostics = new Diagnostics(probeTrace);
    diagnostics->write(stdout);
    delete diagnostics;
    delete probeTrace;
  }
  
  exit(exitCode);
}

int main(int argc, char **argv) 
{
  char *progName = strchr(argv[0], '/') ? (strrchr(argv[0], '/')+1) : argv[0];
  struct Proxy::addressBlock addresses;
  ProbePolicy *policy = new TcpAckPolicy();
  int socketBufferSizeKBytes = -1;
  int receiveBufferSizeBytes = -1;
  int sendBufferSizeBytes = -1;
  char *networkDevice = "eth0";
  bool runAsSender = false;
  bool runAsReceiver = false;
  bool startSenderAndReceiver = true;
  bool notifyParent = false;
  int mtu = 1500;

  memset(&addresses, 0, sizeof(addresses));
  proxy = new TunProxy();
  packetTraceName[0] = 0;
  kernelTraceName[0] = 0;

  /* Parse command line options */

  struct option longOptions[] = {
    { "version",          no_argument,       NULL,   1 },
    { "sender",           required_argument, NULL,   2 },
    { "receiver",         required_argument, NULL,   3 }, 
    { "virtual-sender",   required_argument, NULL,   4 },
    { "virtual-receiver", required_argument, NULL,   5 },
    { "message-file",     required_argument, NULL,   6 },
    { "notify-parent",    no_argument,       NULL,   7 },
    { "transfer-time",    required_argument, NULL,   8 }, /* Only honoured by the sender. Overwrites -z */
    { "verbose",          no_argument,       NULL, 'v' },
    { "transfer-size",    required_argument, NULL, 'z' },
    { "mode",             required_argument, NULL, 'm' },
    { "force-mtu",        required_argument, NULL, 'M' },
    { "interface",        required_argument, NULL, 'i' },
    { "probe-type",       required_argument, NULL, 'p' },
    { "capture-method",   required_argument, NULL, 'c' },
    { NULL, 0, NULL, 0 }
  };

  int opt;
  opterr = 0;
  while ((opt = getopt_long(argc, argv, "o:k:m:M:c:l:i:vz:b:p:t:", longOptions, NULL)) != EOF) {
    switch (opt) {
    case 1 :
      fprintf(stderr, "%s %s\n", PACKAGE, VERSION);
      return 1;
    case 2 :
      if (!parseAddressAndPort(optarg, &addresses.senderAddress))
        panic(ERR_PARAM, "Cannot parse sender address: '%s'", optarg);
      break;
    case 3 :
      if (!parseAddressAndPort(optarg, &addresses.receiverAddress))
        panic(ERR_PARAM, "Cannot parse receiver address: '%s'", optarg);
      break;
    case 4 :
      if (!parseAddressAndPort(optarg, &addresses.receiverPeerAddress))
        panic(ERR_PARAM, "Cannot parse virtual sender address: '%s'", optarg);
      break;
    case 5 :
      if (!parseAddressAndPort(optarg, &addresses.senderPeerAddress))
        panic(ERR_PARAM, "Cannot parse virtual receiver address: '%s'", optarg);
      break;
    case 6 :
      messageFileName = optarg;
      break;
    case 7 :
      notifyParent = true;
      break;
    case 8 :
      transferTimeSec = atoi(optarg);
      break;
    case 'v':
      setVerbosity(1);
      break;
    case 'z':
      transferSizeKB = atoi(optarg);
      break;
    case 'm':
      if (!strcmp(optarg, "standalone"))
        startSenderAndReceiver = false;
      else if (!strcmp(optarg, "sender")) 
        runAsSender = true;
      else if (!strcmp(optarg, "receiver"))
        runAsReceiver = true;
      else if (!strcmp(optarg, "flow"))
        startSenderAndReceiver = true;
      else
        panic(ERR_PARAM, "Unknown mode: '%s'", optarg);
      break;
    case 'l':
      if (!parseAddressAndPort(optarg, &addresses.localAddress))
        panic(ERR_PARAM, "Cannot parse local address: '%s'", optarg);
      break;
    case 'o':
      if (optarg && strcmp(optarg, "auto"))
        strncpy(packetTraceName, optarg, sizeof(packetTraceName));
      if (!packetTrace)
        packetTrace = new MonarchPacketTrace();
      break;
    case 'k':
      if (optarg && strcmp(optarg, "auto"))
        strncpy(kernelTraceName, optarg, sizeof(kernelTraceName));
      if (!kernelTrace)
        kernelTrace = new KernelTrace();
      break;
    case 'M':
      mtu = atoi(optarg);
      break;
    case 'i':
      networkDevice = optarg;
      break;
    case 'b':
      socketBufferSizeKBytes = atoi(optarg);
      receiveBufferSizeBytes = atoi(optarg) * KB;
      sendBufferSizeBytes = atoi(optarg) * KB;
      break;
    case 'p':
      delete policy;
      if (!strcmp(optarg, "tcp-ack"))
        policy = new TcpAckPolicy();
      else if (!strcmp(optarg, "icmp-echo"))
        policy = new IcmpEchoRequestPolicy();
      else if (!strcmp(optarg, "icmp-timestamp"))
        policy = new IcmpTimestampRequestPolicy();
      else if (!strcmp(optarg, "udp-data"))
        policy = new UdpDataPolicy();
      else 
        panic(ERR_PARAM, "Unknown probe type: '%s'", optarg);
      break;
    case 't':
      if (packetTrace)
        delete packetTrace;
      if (!strcmp(optarg, "mout"))
        packetTrace = new MonarchPacketTrace();
#ifdef HAVE_PCAP
      else if (!strcmp(optarg, "pcap"))
        packetTrace = new PcapPacketTrace();
#endif
      else 
        panic(ERR_PARAM, "Unknown trace type: '%s'", optarg);
      break;
    case 'c':
      delete proxy;
      if (!strcmp(optarg, "tun"))
        proxy = new TunProxy();
#ifdef HAVE_IPQ
      else if (!strcmp(optarg, "netfilter"))
        proxy = new NetfilterProxy();
#endif
      else if (!strcmp(optarg, "planetlab"))
        proxy = new PlanetLabProxy();
      else 
        panic(ERR_PARAM, "Unknown capture method: '%s'", optarg);
      break;
    default:
      usage(progName);
    }
  }

  if (optind >= argc)
    usage(progName);
    

  if(addresses.senderAddress.sin_family && addresses.receiverAddress.sin_family) {
    if(addresses.senderAddress.sin_addr.s_addr != addresses.receiverAddress.sin_addr.s_addr)
      panic(ERR_PARAM, "Sender and receiver IP address must be the same!");
  }
  else if(!addresses.senderAddress.sin_family && addresses.receiverAddress.sin_family) {
    panic(ERR_PARAM, "--receiver cannot be used without --sender");
  }

  if(transferTimeSec != -1){
    // If -z and --transfer-time are specified, warn
    if(transferSizeKB != -1){
       fprintf(stderr, "WARNING: --transfer-time and -z specified. -z will be ignored!\n");
    }
    transferSizeKB = 10 *KB; // Set it to a high value (10 MB)
  }
    

  /* Determine local and remote IP addresses */
    
  char *remoteHostArg = argv[optind];
  if (!parseAddressAndPort(remoteHostArg, &addresses.remoteAddress))
    panic(ERR_PARAM, "Cannot parse remote address: '%s'", argv[optind]);
  if (!addresses.remoteAddress.sin_port)
    addresses.remoteAddress.sin_port = htons(10334);

  addresses.localAddress.sin_family = AF_INET;
  if (!addresses.localAddress.sin_addr.s_addr)
    addresses.localAddress.sin_addr.s_addr = getIP(networkDevice);
  if (!addresses.localAddress.sin_port)
    addresses.localAddress.sin_port = htons(10334);

  /* Initialize trace files, if any are requested */

  const char *user = getenv("USER") ? getenv("USER") : "unknown";
  char host[1000];
  if (getenv("HOSTNAME")) {
    strncpy(host, getenv("HOSTNAME"), sizeof(host));
  } else {
    FILE *khostname = fopen("/proc/sys/kernel/hostname", "r");
    if (khostname) {
      fgets(host, sizeof(host), khostname);
      strtok(host, "\r\n");
      fclose(khostname);
    } else {
      strcpy(host, "localhost");
    }
  }
  
  time_t now = time(0);
  char *timestr = ctime(&now);
  strtok(timestr, "\r\n");
  char cmdline[1024];
  strcpy(cmdline, argv[0]);
  for (int i=1; i<argc; i++) {
    strcat(cmdline, " ");
    strcat(cmdline, argv[i]);
  }

  if (packetTrace) {
    if (!packetTraceName[0]) {
      char buf[200];
      strncpy(buf, remoteHostArg, sizeof(buf));
      strtok(buf, ":\r\n");
      sprintf(packetTraceName, "%s_%s_%d%s", host, buf, (int)now, packetTrace->getDefaultExtension());
    }

    verbose(1, "Initializing %s packet trace '%s'", packetTrace->getName(), packetTraceName);
    packetTrace->setFileName(packetTraceName);
    packetTrace->init();
    
    packetTrace->addComment("%s %s trace collected by %s@%s on %s", PACKAGE, VERSION, user, host, timestr);
    packetTrace->addComment("Command line: %s", cmdline);
  }
  
  if (kernelTrace) {
    char buf[1024];
    if (!kernelTraceName[0]) {
      strncpy(buf, remoteHostArg, sizeof(buf));
      strtok(buf, ":\r\n");
      sprintf(kernelTraceName, "%s_%s_%d", host, buf, (int)now);
    }
    
    if (runAsSender)
      strcat(kernelTraceName, ".send");
    if (runAsReceiver)
      strcat(kernelTraceName, ".recv");
      
    kernelTrace->setFileName(kernelTraceName);

    kernelTrace->addComment("%s %s trace collected by %s@%s on %s", PACKAGE, VERSION, user, host, timestr);
    kernelTrace->addComment("Command line: %s", cmdline);
    kernelTrace->addComment("");
    kernelTrace->addComment("--- Protocol stack configuration");
    for (unsigned i=0; i<sizeof(procEntries)/sizeof(char*); i++) {
      char sbuf[200];
      strcpy(sbuf, "--");
      
      FILE *procfile = fopen(procEntries[i], "r");
      if (procfile) {
        fgets(sbuf, sizeof(sbuf), procfile);
        strtok(sbuf, "\r\n");
        fclose(procfile);
      }
      
      kernelTrace->addComment("%s: %s", procEntries[i], sbuf);
    }

    kernelTrace->addComment("");
    kernelTrace->addComment("--- File format");
    kernelTrace->addComment("01: Time");
    for (unsigned i=0; i<sizeof(columnName)/sizeof(char*); i++)
      kernelTrace->addComment("%02d: %s", i+2, columnName[i]);
    kernelTrace->addComment("");
  }

  /* Run as a sender, receiver, or proxy */
    
  if (runAsSender) {
    Sender *sender = new Sender(&addresses.localAddress, &addresses.remoteAddress);
    if (transferSizeKB >= 0)
      sender->setTransferSize(transferSizeKB * KB);
    if(transferTimeSec >= 0) 
      sender->setTransferTime(transferTimeSec);
    if (sendBufferSizeBytes >= 0)
      sender->setSendBufferSize(sendBufferSizeBytes);
    if (messageFileName)
      sender->setMessageFile(messageFileName);
    if (kernelTrace)
      sender->setKernelTrace(kernelTrace);
    sender->run();
    delete sender;
    if (kernelTrace)
      kernelTrace->finalize();
  } else if (runAsReceiver) {
    Receiver *receiver = new Receiver(&addresses.localAddress, &addresses.remoteAddress);
    if (receiveBufferSizeBytes >= 0)
      receiver->setReceiveBufferSize(receiveBufferSizeBytes);
    receiver->init(); // Must not be called before setReceiveBufferSize() !
    if (kernelTrace)
      receiver->setKernelTrace(kernelTrace);
    if (notifyParent)
      kill(getppid(), SIGUSR1);
    receiver->run();
    delete receiver;
    if (kernelTrace)
      kernelTrace->finalize();
  } else {
    if (packetTrace) {
      packetTrace->addComment("Capture method: %s", proxy->getName());
      proxy->setPacketTrace(packetTrace);
    }
    if (mtu != 1500)
      proxy->setMTU(mtu);
    
    probeTrace = new ProbeTrace();
    policy->setProbeTrace(probeTrace);
    proxy->setProbePolicy(policy);
    
    verbose(1, "Initializing %s proxy", proxy->getName());
    proxy->init(&addresses, networkDevice);
    
    signal(SIGTERM, handleTermination);
    signal(SIGHUP, handleTermination);
    signal(SIGINT, handleTermination);
    signal(SIGABRT, handleTermination);
    signal(SIGCHLD, handleChildExit);

    if (startSenderAndReceiver) {
      struct sockaddr_in senderLocal, senderRemote, receiverLocal, receiverRemote;
      verbose(1, "Launching internal TCP sender and receiver");
      proxy->getAddresses(&senderLocal, &senderRemote, &receiverLocal, &receiverRemote);
      forkSenderAndReceiver(argv[0], &senderLocal, &senderRemote, &receiverLocal, &receiverRemote, socketBufferSizeKBytes);
    } else {
      verbose(1, "Running in standalone mode");
    }
    
    proxy->run();
    assert(false);
  }

  return 0;
}
