#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>

#include "packet.h"

Packet::Packet()
{
  this->len = 0;
}

Packet::Packet(unsigned char *data, int lenArg)
{
  assert((0<=lenArg) && (lenArg<=(int)sizeof(payload)));
  if (lenArg > 0)
    memcpy(&payload, data, lenArg);
  this->len = lenArg;
}

bool Packet::isTcpPacket()
{
  return getIpProtocolNumber() == IPPROTO_TCP;
}

bool Packet::isUdpPacket()
{
  return getIpProtocolNumber() == IPPROTO_UDP;
}

int Packet::getIpLength()
{
  struct IpHeader *ip = (struct IpHeader*) &payload;
  return ntohs(ip->ip_len);
}

struct TcpHeader *Packet::findTcpHeader()
{
  assert(isTcpPacket());
  struct IpHeader *ip = (struct IpHeader*) &payload;
  int ipHeaderLen = 4*(ip->ip_vhl & 0xF);
  assert(ipHeaderLen < len);

  return (struct TcpHeader*)&payload[ipHeaderLen];
}

struct UdpHeader *Packet::findUdpHeader()
{
  assert(isUdpPacket());
  struct IpHeader *ip = (struct IpHeader*) &payload;
  int ipHeaderLen = 4*(ip->ip_vhl & 0xF);
  assert(ipHeaderLen < len);

  return (struct UdpHeader*)&payload[ipHeaderLen];
}

Packet::Packet(Packet *original)
{
  if (original->len > 0)
    memcpy(&payload[0], &original->payload[0], original->len);
    
  this->len = original->len;
}

void Packet::copyFrom(unsigned char *data, int lenBytes)
{
  assert(data && (lenBytes>0) && (lenBytes<=(int)sizeof(payload)));
  memcpy(payload, data, lenBytes);
  len = lenBytes;
}

bool Packet::isSource(struct sockaddr_in *addr)
{
  assert(addr->sin_family == AF_INET);
  return ((ntohl(addr->sin_addr.s_addr) == getSource()) && (ntohs(addr->sin_port) == getSourcePort()));
}

uint16 Packet::getSourcePort() 
{ 
  if (isTcpPacket())
    return ntohs(findTcpHeader()->tcp_sport);
  if (isUdpPacket())
    return ntohs(findUdpHeader()->sourcePort);
  return 0; 
}

uint16 Packet::getDestinationPort() 
{ 
  if (isTcpPacket())
    return ntohs(findTcpHeader()->tcp_dport);
  if (isUdpPacket())
    return ntohs(findUdpHeader()->destPort);
  return 0; 
}

bool Packet::isDestination(struct sockaddr_in *addr)
{
  assert(addr->sin_family == AF_INET);
  return ((ntohl(addr->sin_addr.s_addr) == getDestination()) && (ntohs(addr->sin_port) == getDestinationPort()));
}

void Packet::dump(FILE *stream)
{
  int i, off = 0;
  while (off < len) {
    fprintf(stream, "%04X ", off);
    for (i=0; i<16; i++) {
      if ((i+off) < len)
        fprintf(stream, "%02X ", payload[i+off]);
      else
        fprintf(stream, "   ");
    }
    off += 16;
    fprintf(stream, "\n");
  }
}

uint16 Packet::checksum(uint8 *p1, int size1, uint8 *p2, int size2)
{
  assert(p1 && (size1>=0) && (size2>=0));
  uint32 sum=0;

  for (int k=0; k<2; k++) {
    uint8 *lpacket = (k==0) ? p1 : p2;
    if (lpacket) {
      int size = (k==0) ? size1 : size2;
      for (int i=0; i<size; i+=2) {
        unsigned short word16;
        if ((i+1)<size)
          word16 = (((uint32)lpacket[i])<<8)+lpacket[i+1];
        else
          word16 = (((uint32)lpacket[i])<<8);
        sum += (uint32) word16;
      }
    }
  }
    
  while (sum>>16)
    sum = (sum & 0xFFFF) + (sum>>16);
    
  return (~sum)&0xFFFF;
}

void Packet::recalculateChecksum()
{
  struct IpHeader *hip = (struct IpHeader*) payload;
  int ihl = (hip->ip_vhl&0xF)*4;
  assert((0<=ihl) && ((int)ihl<=len));
  hip->ip_xsum = 0;
  hip->ip_xsum = htons(checksum((uint8*) hip, 0, (uint8*) hip, ihl));
  
  /* TCP and UDP packets have a second checksum */
  
  if (hip->ip_proto == IPPROTO_TCP) {
    uint8 *lpacket = (uint8*) hip;
    struct TcpHeader *htcp = (struct TcpHeader*)(&lpacket[ihl]);
    htcp->tcp_xsum = 0;
    struct tcpPseudoHdr pseudoHdr = { hip->ip_src, hip->ip_dst, 0, hip->ip_proto, htons(len - ihl) };
    htcp->tcp_xsum = htons(checksum((uint8*)&pseudoHdr, sizeof(struct tcpPseudoHdr), (uint8*)htcp, len-ihl));
  } else if (hip->ip_proto == IPPROTO_UDP) {
    uint8 *lpacket = (uint8*) hip;
    struct UdpHeader *hudp = (struct UdpHeader*)(&lpacket[ihl]);
    hudp->checksum = 0;
    struct tcpPseudoHdr pseudoHdr = { hip->ip_src, hip->ip_dst, 0, hip->ip_proto, htons(len - ihl) };
    hudp->checksum = htons(checksum((uint8*)&pseudoHdr, sizeof(struct tcpPseudoHdr), (uint8*)hudp, len-ihl));
  }
}

void Packet::setSourceAndDestination(struct sockaddr_in *source, struct sockaddr_in *destination)
{
  assert(source->sin_family == AF_INET);
  assert(destination->sin_family == AF_INET);
  
  ((struct IpHeader*)&payload[0])->ip_src = source->sin_addr.s_addr;
  ((struct IpHeader*)&payload[0])->ip_dst = destination->sin_addr.s_addr;
  
  if (isTcpPacket()) {
    struct TcpHeader *tcpHeader = findTcpHeader();
    tcpHeader->tcp_sport = source->sin_port;
    tcpHeader->tcp_dport = destination->sin_port;
  } else if (isUdpPacket()) {
    struct UdpHeader *udpHeader = findUdpHeader();
    udpHeader->sourcePort = source->sin_port;
    udpHeader->destPort = destination->sin_port;
  }
  
  recalculateChecksum();
}

int Packet::getIpHeaderLength()
{
  struct IpHeader *iph = (struct IpHeader*)&payload[0];
  return (iph->ip_vhl & 0xF)<<2;
}

int Packet::getTcpHeaderLength()
{
  if (!isTcpPacket())
    return 0;
    
  return (findTcpHeader()->tcp_hl >> 4) << 2;
}

uint8 *Packet::findTcpOption(int type)
{
  if (!isTcpPacket())
    return NULL;
  
  int bytesRemaining = getTcpHeaderLength() - 5*4;
  uint8 *ptr = &(((uint8*)findTcpHeader())[20]);
  while ((bytesRemaining>0) && *ptr) {
    if (*ptr == type)
      return ptr;
    int len = ((ptr[0] == 0) || (ptr[0] == 1)) ? 1 : ptr[1];
    assert(len > 0);
    bytesRemaining -= len;
    ptr += len;
  }
  
  return NULL;
}
