#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>

#include "monarch.h"
#include "traces.h"
#include "packet.h"

MonarchPacketTrace::MonarchPacketTrace()
{
  firstBlock = (struct packetBlock*) malloc(sizeof(struct packetBlock));
  assert(firstBlock != NULL);
  
  firstBlock->packetsHere = 0;
  firstBlock->nextBlock = NULL;
  lastBlock = firstBlock;
  firstComment = NULL;
  lastComment = NULL;
}

MonarchPacketTrace::~MonarchPacketTrace()
{
  while (firstBlock != NULL) {
    struct packetBlock *nextBlock = firstBlock->nextBlock;
    free(firstBlock);
    firstBlock = nextBlock;
  }
  
  while (firstComment != NULL) {
    struct commentRecord *nextComment = firstComment->next;
    free(firstComment);
    firstComment = nextComment;
  }
}

void MonarchPacketTrace::addComment(const char *format, ...)
{
  char buffer[1024];
  va_list ap;
  va_start(ap, format);
  vsnprintf(buffer, sizeof(buffer), format, ap);
  struct commentRecord *comment = (struct commentRecord*) malloc(sizeof(commentRecord));
  assert(comment != NULL);
  comment->next = NULL;
  strncpy(comment->text, buffer, sizeof(comment->text));
  
  if (lastComment) 
    lastComment->next = comment;
  else
    firstComment = comment;
  lastComment = comment;
  
  va_end (ap);
}

void MonarchPacketTrace::addPacket(struct timeval *time, Packet *packet)
{
  if (lastBlock->packetsHere >= packetsPerBlock) {
    assert(lastBlock->nextBlock == NULL);
    lastBlock->nextBlock = (struct packetBlock*) malloc(sizeof(struct packetBlock));
    assert(lastBlock->nextBlock != NULL);
    lastBlock = lastBlock->nextBlock;
    lastBlock->packetsHere = 0;
    lastBlock->nextBlock = NULL;
  }
  
  int captureLen = packet->getIpHeaderLength();
  if (packet->isTcpPacket())
    captureLen += packet->getTcpHeaderLength();
  else if (packet->isUdpPacket())
    captureLen += sizeof(struct UdpHeader);
  else
    captureLen = packet->getIpLength();
    
  if (captureLen > maxHeaderLength)
    captureLen = maxHeaderLength;
    
  int thisIndex = lastBlock->packetsHere ++;
  memcpy(&lastBlock->packet[thisIndex].headers, packet->getBuffer(), captureLen);
  lastBlock->packet[thisIndex].headerLen = captureLen;
  lastBlock->packet[thisIndex].time = *time;
}

void MonarchPacketTrace::writePacket(struct packetRecord *packet, FILE *stream)
{
  char buf1[256], buf2[256];
  Packet *cpacket = new Packet(packet->headers, packet->headerLen);
  
  fprintf(stream, "%d.%06d ", (int)packet->time.tv_sec, (int)packet->time.tv_usec);

  if (cpacket->isTcpPacket() || cpacket->isUdpPacket()) {
    fprintf(stream, "%s.%d > %s.%d",
      renderIP(cpacket->getSource(), buf1), cpacket->getSourcePort(), 
      renderIP(cpacket->getDestination(), buf2), cpacket->getDestinationPort()
    );
  } else {
    fprintf(stream, "%s > %s proto %d, id %d", 
      renderIP(cpacket->getSource(), buf1), renderIP(cpacket->getDestination(), buf2),
      cpacket->getIpProtocolNumber(), cpacket->getIpId()
    );
  }
  
  if (cpacket->isTcpPacket()) {
    int bits = cpacket->getControlBits();
    char flags[20];
    int soFar = 0;
    
    if (bits & Packet::FLAG_FIN)
      flags[soFar++] = 'F';
    if (bits & Packet::FLAG_SYN)
      flags[soFar++] = 'S';
    if (bits & Packet::FLAG_RST)
      flags[soFar++] = 'R';
    if (bits & Packet::FLAG_PSH)
      flags[soFar++] = 'P';
    if (bits & Packet::FLAG_ACK)
      flags[soFar++] = 'A';
    if (bits & Packet::FLAG_URG)
      flags[soFar++] = 'U';
    if (!soFar)
      flags[soFar++] = '-';
    flags[soFar] = 0;

    fprintf(stream, " %s seq: %u, ack: %u id: %d, ttl: %d win: %d, urg: %d datalen: %d, optlen: %d",
      flags, cpacket->getTcpSequenceNo(), cpacket->getTcpAckNo(),
      cpacket->getIpId(), cpacket->getTTL(), cpacket->getTcpWindowSize(),
      cpacket->getTcpUrgentPointer(), cpacket->getIpLength() - (cpacket->getIpHeaderLength() + cpacket->getTcpHeaderLength()),
      cpacket->getTcpHeaderLength() - 20
    );

    uint8 *mssOption = cpacket->findTcpOption(Packet::TCP_OPTION_MSS);
    if (mssOption)
      fprintf(stream, " mss: %d", ntohs(*((uint16*)&mssOption[2])));
  } else if (cpacket->isUdpPacket()) {
    fprintf(stream, " udp, datalen: %d, id %d",
      cpacket->getIpLength() - (cpacket->getIpHeaderLength() + sizeof(struct UdpHeader)), cpacket->getIpId()
    );
  }
  
  fprintf(stream, "\n");

  for (int off=0; off < packet->headerLen; ) {  
    fprintf(stream, "    ");
    for (int i=0; i<16; i++) {
      if ((i+off) < packet->headerLen)
        fprintf(stream, "%02X ", packet->headers[i+off]);
      else
        fprintf(stream, "   ");
    }
    off += 16;
    fprintf(stream, "\n");
  }
    
  delete cpacket;
}

void MonarchPacketTrace::finalize()
{
  FILE *stream = fopen(outFileName, "w+");
  if (!stream)
    panic(ERR_FILEIO, "Cannot write trace to '%s'", outFileName);
  
  struct commentRecord *currentComment = firstComment;
  while (currentComment) {
    fprintf(stream, "# %s\n", currentComment->text);
    currentComment = currentComment->next;
  }
  
  struct packetBlock *currentBlock = firstBlock;
  while (currentBlock) {
    for (int i=0; i<currentBlock->packetsHere; i++)
      writePacket(&currentBlock->packet[i], stream);
    currentBlock = currentBlock->nextBlock;
  }
  
  fclose(stream);
}
